/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2011, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */


#include <sys/stat.h>
#include "cc_mgr.h"


/****************************************************/
/**************** private methods *******************/
/****************************************************/
//============================================================================//
void CongestionControlManager::ResetUserOptionsDataBase()
{
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    map<int,CCTIEntryListVec>::iterator ca_ccti_list_it;

    m_user_option_db.m_is_set_flags.clear();

    sw_node_it =
        m_user_option_db.m_sw_cc_setting_db.m_nodes_db.begin();
    while (sw_node_it !=
           m_user_option_db.m_sw_cc_setting_db.m_nodes_db.end()) {
        sw_node_it->second.m_is_set_flags.clear();
        sw_node_it++;
    }
    m_user_option_db.m_sw_cc_setting_db.m_nodes_db.clear();
    m_user_option_db.m_sw_cc_setting_db.m_is_set_flags.clear();

    ca_node_it =
        m_ca_cc_setting_db.m_nodes_db.begin();
    while (ca_node_it !=  m_ca_cc_setting_db.m_nodes_db.end()) {
        ca_node_it->second.m_is_set_flags.clear();
        ca_node_it++;
    }
    m_user_option_db.m_ca_cc_setting_db.m_nodes_db.clear();
    m_user_option_db.m_ca_cc_setting_db.m_is_set_flags.clear();

    ca_ccti_list_it =
        m_user_option_db.m_ca_cc_setting_db.m_default_ccti_list.begin();
    while (ca_ccti_list_it !=
           m_user_option_db.m_ca_cc_setting_db.m_default_ccti_list.end()) {
        ca_ccti_list_it->second.clear();
        ca_ccti_list_it++;
    }
    m_user_option_db.m_ca_cc_setting_db.m_default_ccti_list[0] =
        CCTIEntryListVec();
}


//============================================================================//
/***************/
vector<string> CongestionControlManager::Split(const string& s)
{
    istringstream iss(s);
    return vector<string>(istream_iterator<string>(iss), istream_iterator<string>());
}

/***************/
bool CongestionControlManager::CheckStr(const string& str_value)
{
    std::string aux_str_value(str_value);
    if (0 == aux_str_value.find("0x")) {
        aux_str_value.erase(0,2);
    }
    if (aux_str_value.find_first_not_of("0") != str_value.npos) {
        return false;
    }
    return true;
}

bool CongestionControlManager::CheckAndSetUserOptionFlag(const string& key,
        std::map<string, bool>& user_options_flags)
{
    bool rc = true;
    if (user_options_flags.find(key) != user_options_flags.end()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "The following parameter: %s, was already set\n",
                key.c_str());
        rc = false;
    }
    user_options_flags[key] = true;
    return rc;
}

bool CongestionControlManager::CheckEnableConflict(bool enable,
        const map<string, bool> &is_set_flags,
        const string &key_str_type)
{
    if (enable && m_user_option_db.m_enable) {
        return false;
    }
    stringstream ss;
    bool is_conflict_key_set = false;
    map<string, bool>::const_iterator is_set_flags_it = is_set_flags.begin();
    while (is_set_flags_it != is_set_flags.end()) {
        if (is_set_flags_it->first != string("enable") &&
            is_set_flags_it->first != string("cc_key") &&
            is_set_flags_it->first != string("max_errors") &&
            is_set_flags_it->first != string("error_window") &&
            is_set_flags_it->first != string("cc_statistics_cycle") &&
            is_set_flags_it->first != string("cc_log_file_name") &&
            is_set_flags_it->first != string("cc_log_file_size")) {
            is_conflict_key_set = true;
            ss << is_set_flags_it->first << ", ";
        }
        is_set_flags_it++;
    }
    if (is_conflict_key_set) {
        // Found keys that conflict with enable
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "\"enable false\" option conflicts with the following %s parameters:"
                " %s. \"enable false\" will override them\n",
                key_str_type.c_str(), ss.str().c_str());
        return true;
    }
    return false;
}

/***************/
int CongestionControlManager::SetGeneralUserOptions(const string &key, const string &str_value)
{
    int rc = 0;
    u_int64_t value = strtoull(str_value.c_str(),NULL,0);

    switch (m_general_user_options[key]) {
    case ENABLE:
        if (str_value != string("FALSE") && str_value != string("TRUE")) {
            return 1;
        }
        m_user_option_db.m_enable = (str_value == string("TRUE")) ? true : false;
        break;
    case CC_KEY:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        m_user_option_db.m_cc_key = value;
        break;
    case NUM_HOSTS:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        m_user_option_db.m_num_hosts = (u_int32_t) value;
        break;
    case MAX_ERRORS:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        m_user_option_db.m_max_errors = value;
        break;
    case ERROR_WINDOW:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        m_user_option_db.m_error_window = value;
        break;
    case CC_STATISTICS_CYCLE:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        m_user_option_db.m_cc_statistics_cycle = value;
        break;
    case CC_LOG_FILE_NAME:
        m_user_option_db.m_cc_log_file_name = str_value;
        break;
    case CC_LOG_FILE_SIZE:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        m_user_option_db.m_cc_log_file_size = value;
        break;
    default:
        return 1;
        break;
    }

    CheckAndSetUserOptionFlag(key, m_user_option_db.m_is_set_flags);
    CheckEnableConflict(m_user_option_db.m_enable, m_user_option_db.m_is_set_flags, "General");
    return rc;
}

int CongestionControlManager::SetCAUserOption(const string &key, const string &str_value,
        struct CC_CACongestionSetting& cc_ca_congestion_setting,
        CCTIEntryListVec& ccti_list,
        map<string, bool>& is_set_flags,
        bool& enable)
{
    int rc = 0;
    u_int32_t value = strtoull(str_value.c_str(),NULL,0);
    vector<string> cct_str;

    switch (m_ca_user_options[key]) {
    case CA_ENABLE:
        if (str_value != string("FALSE") && str_value != string("TRUE")) {
            return 1;
        }
        enable = (str_value == string("TRUE")) ? true : false;
        break;
    case PORT_CONTROL:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        cc_ca_congestion_setting.Port_Control = value;
        break;
    case CA_CONTROL_MAP:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        cc_ca_congestion_setting.Control_Map = value;
        break;
    case CCTI_TIMER:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
            cc_ca_congestion_setting.CACongestionEntryList.
                CACongestionEntryListElement[sl].CCTI_Timer = value;
        }
        break;
    case CCTI_INCR:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
            cc_ca_congestion_setting.CACongestionEntryList.
                CACongestionEntryListElement[sl].CCTI_Increase = value;
        }
        break;
    case TRIGGER_THRESHOLD:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
            cc_ca_congestion_setting.CACongestionEntryList.
                CACongestionEntryListElement[sl].Trigger_Threshold = value;
        }
        break;
    case CCTI_MIN:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
            cc_ca_congestion_setting.CACongestionEntryList.
                CACongestionEntryListElement[sl].CCTI_Min = value;
        }
        break;
    case CCT:
        {
            string tmp_str_value = str_value;
            replace_if(tmp_str_value.begin(), tmp_str_value.end(),
                       std::bind2nd(std::equal_to<char>(),','), ' ');
            cct_str = Split(tmp_str_value);
            if (!cct_str.size()) {
                break;
            }
            if (cct_str[0] == string("MONITOR")) {
                // Create cct for MONITOR mode
                unsigned int num_blocks = 2;
                ccti_list.clear();
                for (unsigned int block = 0; block < num_blocks; block++) {
                    struct CCTI_Entry_List ccti_entry_list;
                    CLEAR_STRUCT(ccti_entry_list);
                    ccti_list.push_back(ccti_entry_list);
                }
                break;
            }
            vector<u_int16_t> cct;
            for (unsigned int idx = 0; idx < cct_str.size(); idx++) {
                cct.push_back((u_int16_t) strtoul(cct_str[idx].c_str(),NULL,0));
                if (!cct[cct.size()-1] && !CheckStr(cct_str[idx])) {
                    return 1;
                }
            }
            if (cct.size() == 1 && cct[0] == 0) {
                // Default, no thing to do.
                break;
            }

            unsigned int num_blocks = (cct.size() + ccti_entry_list_elements_num - 1) /
                                       ccti_entry_list_elements_num;
            num_blocks = (num_blocks <= 2) ? 2 : num_blocks;

            ccti_list.clear();
            for (unsigned int block = 0; block < num_blocks; block++) {
                struct CCTI_Entry_List ccti_entry_list;
                CLEAR_STRUCT(ccti_entry_list);
                ccti_list.push_back(ccti_entry_list);
            }
            for (unsigned int idx = 0; idx < cct.size(); idx++) {
                ccti_list[idx / ccti_entry_list_elements_num].
                    CCTI_Entry_ListElement[idx % ccti_entry_list_elements_num].CCT_Multiplier =
                (cct[idx] & 0x3fff);
                ccti_list[idx / ccti_entry_list_elements_num].
                    CCTI_Entry_ListElement[idx % ccti_entry_list_elements_num].CCT_Shift =
                (cct[idx] & 0xc000) >> 14;
            }
            if (cct.size() == num_blocks * ccti_entry_list_elements_num) {
                break;
            }
            for (unsigned int idx = cct.size(); idx < num_blocks * ccti_entry_list_elements_num; idx++) {
                ccti_list[idx / ccti_entry_list_elements_num].
                    CCTI_Entry_ListElement[idx % ccti_entry_list_elements_num].CCT_Multiplier =
                (cct[cct.size() - 1] & 0x3fff);
                ccti_list[idx / ccti_entry_list_elements_num].
                    CCTI_Entry_ListElement[idx % ccti_entry_list_elements_num].CCT_Shift =
                (cct[cct.size() - 1] & 0xc000) >> 14;
            }
        }
        break;
    default:
        return 1;
        break;
    }

    CheckAndSetUserOptionFlag(key, is_set_flags);
    CheckEnableConflict(enable, is_set_flags, "CA");
    return rc;
}

int CongestionControlManager::SetSWUserOption(const string &key, const string &str_value,
        struct CC_SwitchCongestionSetting& cc_sw_congestion_setting,
        map<string, bool>& is_set_flags,
        bool& enable)
{
    int rc = 0;
    u_int32_t value = strtoull(str_value.c_str(),NULL,0);

    switch (m_sw_user_options[key]) {
    case SW_ENABLE:
        if (str_value != string("FALSE") && str_value != string("TRUE")) {
            return 1;
        }
        enable = (str_value == string("TRUE")) ? true : false;
        break;
    case THRESHOLD:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        cc_sw_congestion_setting.Threshold = value;
        break;
    case PACKET_SIZE:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        cc_sw_congestion_setting.Packet_Size = value / 64 + (value % 64 != 0); // Convert from bytes to credits
        break;
    case MARKING_RATE:
        if (!value && !CheckStr(str_value)) {
            return 1;
        }
        cc_sw_congestion_setting.Marking_Rate = value;
        break;
    default:
        return 1;
        break;
    }

    CheckAndSetUserOptionFlag(key, is_set_flags);
    CheckEnableConflict(enable, is_set_flags, "Switch");
    return rc;
}

/***************/
bool CongestionControlManager::IsFileExists(const char *file_name)
{
    OSM_CC_LOG_ENTER(this->m_p_osm_log);

    bool rc;

    struct stat file_info;
    memset(&file_info, 0, sizeof(file_info));

    int rc_stat = stat(file_name,
            &file_info);
    if (rc_stat == 0) {
        // We were able to get the file attributes
        // so the file obviously exists
        rc = true;
    } else {    // Not exists or not permission
        rc = false;
    }

    OSM_CC_LOG_RETURN(this->m_p_osm_log, rc);
}

/***************/
int CongestionControlManager::ParseInputFile()
{
    int rc = 0;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;

    ResetUserOptionsDataBase();

    u_int8_t node_type;
    map<string, USER_OPTIONS_STR>* p_user_options = NULL;

    // Check if configuration file exists
    if (this->IsFileExists(this->m_conf_file_name.c_str()) == false) {
        OSM_CC_LOG(this->m_p_osm_log, OSM_LOG_INFO, "Configuration file not exists, using default values\n");
        fprintf(stdout, "CC Manager - Configuration file not exists, using default values\n");
        return rc;
    }

    ifstream input_file(m_conf_file_name.c_str());
    if (!input_file) {
        OSM_CC_LOG(this->m_p_osm_log, OSM_LOG_INFO, "Parse of configuration file failed, using default values\n");
        fprintf(stdout, "CC Manager - Parse of configuration file failed, using default values\n");
        return rc;
    }
    string lineread;
    while (getline(input_file, lineread)) { // Read line by line
        //...lineread contains the line read
        if (lineread == "" || lineread[0] == '#') {
            //skip empty lines
            continue;
        }
        replace_if(lineread.begin(), lineread.end(),
                   bind2nd(std::equal_to<char>(),'='), ' ');
        vector<string> v = Split(lineread);
        if (!v.size()) {
            continue;
        }
        u_int64_t port_guid = strtoull(v[0].c_str(),NULL,0);
        if (!port_guid) {
            for (unsigned int key_idx = 0;
                  key_idx < v.size() - 1;
                  key_idx = key_idx + 2) {
                rc = 1;
                string node_type_str;
                if (m_general_user_options.find(v[key_idx]) !=
                    m_general_user_options.end()) {
                    node_type_str = "General";
                    rc = SetGeneralUserOptions(v[key_idx], v[key_idx+1]);
                }
                else if (m_ca_user_options.find(v[key_idx]) !=
                         m_ca_user_options.end()) {
                    node_type_str = "CA";
                    rc = SetCAUserOption(v[key_idx], v[key_idx+1],
                                         m_user_option_db.m_ca_cc_setting_db.m_default_params,
                                         m_user_option_db.m_ca_cc_setting_db.
                                         m_default_ccti_list[CC_GENERAL_DEFAULT_CCTI_LIST],
                                         m_user_option_db.m_ca_cc_setting_db.m_is_set_flags,
                                         m_user_option_db.m_enable);
                }
                else if (m_sw_user_options.find(v[key_idx]) != m_sw_user_options.end()) {
                    node_type_str = "Switch";
                    rc = SetSWUserOption(v[key_idx], v[key_idx+1],
                                         m_user_option_db.m_sw_cc_setting_db.m_default_params,
                                         m_user_option_db.m_sw_cc_setting_db.m_is_set_flags,
                                         m_user_option_db.m_enable);
                }
                else {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                            "Parameter %s is unknown\n",v[key_idx].c_str());
                    continue;
                }
                if (!rc) {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                            "Default %s parameter was set: %s = %s\n",
                            node_type_str.c_str(), v[key_idx].c_str(), v[key_idx + 1].c_str());
                }
                else {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                            "Default %s parameter was NOT set: %s = %s\n",
                            node_type_str.c_str(), v[key_idx].c_str(), v[key_idx + 1].c_str());
                }
            }
        }
        else if (port_guid && !CheckNodeType(port_guid, node_type)) {
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Given GUID doesn't exist (GUID : " U64H_FMT ")\n", port_guid);
        }
        else if (port_guid && node_type == IB_NODE_TYPE_ROUTER) {
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Given GUID is a router. Congestion Control Manager doesn't support"
                    " configuring a router (GUID : " U64H_FMT ")\n", port_guid);
        }
        else {
            if (node_type == IB_NODE_TYPE_CA) {
                if ((ca_node_it = m_user_option_db.m_ca_cc_setting_db.m_nodes_db.find(port_guid)) ==
                    m_user_option_db.m_ca_cc_setting_db.m_nodes_db.end()) {
                    CACCSettingDataBaseNodeEntry new_ca_node(
                        port_guid, m_ca_cc_setting_db.m_default_params);
                    m_user_option_db.m_ca_cc_setting_db.m_nodes_db.insert(
                      pair<u_int64_t, CACCSettingDataBaseNodeEntry>(port_guid, new_ca_node));
                    ca_node_it = m_user_option_db.m_ca_cc_setting_db.m_nodes_db.find(port_guid);
                }
            }
            else {
                if ((sw_node_it = m_user_option_db.m_sw_cc_setting_db.m_nodes_db.find(port_guid)) ==
                    m_user_option_db.m_sw_cc_setting_db.m_nodes_db.end()) {
                    SWCCSettingDataBaseNodeEntry new_sw_node(port_guid, m_sw_cc_setting_db.m_default_params);
                    m_user_option_db.m_sw_cc_setting_db.m_nodes_db.insert(
                      pair<u_int64_t, SWCCSettingDataBaseNodeEntry>(port_guid, new_sw_node));
                    sw_node_it = m_user_option_db.m_sw_cc_setting_db.m_nodes_db.find(port_guid);
                }
            }

            p_user_options = (node_type == IB_NODE_TYPE_CA) ? &m_ca_user_options : &m_sw_user_options;
            string node_type_str = (node_type == IB_NODE_TYPE_CA) ? "CA" : "Switch";
            for (unsigned int key_idx = 1; key_idx < v.size() - 1; key_idx = key_idx + 2) {
                if (p_user_options->find(v[key_idx]) == p_user_options->end()) {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                            "Parameter %s is unknown\n", v[key_idx].c_str());
                    continue;
                }
                if (node_type == IB_NODE_TYPE_CA) {
                    rc = SetCAUserOption(v[key_idx], v[key_idx+1],
                                         ca_node_it->second.m_node_params,
                                         ca_node_it->second.m_node_ccti_list,
                                         ca_node_it->second.m_is_set_flags,
                                         ca_node_it->second.m_enable);
                }
                else {
                    rc = SetSWUserOption(v[key_idx], v[key_idx+1],
                                         sw_node_it->second.m_node_params,
                                         sw_node_it->second.m_is_set_flags,
                                         sw_node_it->second.m_enable);
                }
                if (!rc) {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                            "GUID: " U64H_FMT ". %s parameter was set: %s = %s\n",
                            port_guid, node_type_str.c_str(), v[key_idx].c_str(), v[key_idx + 1].c_str());
                }
                else {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                            "GUID: " U64H_FMT ". %s parameter was NOT set: %s = %s. "
                            "Value is illegal\n",
                            port_guid, node_type_str.c_str(), v[key_idx].c_str(),
                            v[key_idx + 1].c_str());
                }
            }
        }
    }
    return rc;
}


//============================================================================//
/***************/
void CongestionControlManager::SetGeneralOptionsFromUser(const string &key)
{
    switch (m_general_user_options[key]) {
    case ENABLE:
        m_enable = m_user_option_db.m_enable;
        break;
    case CC_KEY:
        m_cc_key = m_user_option_db.m_cc_key;
        break;
    case NUM_HOSTS:
        m_num_hosts = m_user_option_db.m_num_hosts;
        break;
    case MAX_ERRORS:
        m_max_errors = m_user_option_db.m_max_errors;
        break;
    case ERROR_WINDOW:
        m_error_window = m_user_option_db.m_error_window;
        break;
    case CC_STATISTICS_CYCLE:
        m_cc_statistics_cycle = m_user_option_db.m_cc_statistics_cycle;
        break;
    case CC_LOG_FILE_NAME:
        m_cc_log_file_name = m_user_option_db.m_cc_log_file_name;
        break;
    case CC_LOG_FILE_SIZE:
        m_cc_log_file_size = m_user_option_db.m_cc_log_file_size;
        break;
    default:
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "Enterted function with key = %s\n", key.c_str());
        break;
    }
}

void CongestionControlManager::SetSWOptionsFromUser(const string &key,
        const struct CC_SwitchCongestionSetting src_params,
        struct CC_SwitchCongestionSetting& dest_params,
        bool& enable)
{
    switch (m_sw_user_options[key]) {
    case SW_ENABLE:
        enable = true;
        break;
    case THRESHOLD:
        dest_params.Threshold = src_params.Threshold;
        break;
    case PACKET_SIZE:
        dest_params.Packet_Size = src_params.Packet_Size;
        break;
    case MARKING_RATE:
        dest_params.Marking_Rate = src_params.Marking_Rate;
        break;
    default:
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "Enterted function with key = %s\n", key.c_str());
        break;
    }
}

void CongestionControlManager::SetCAOptionsFromUser(const string &key,
        const struct CC_CACongestionSetting &src_params,
        const CCTIEntryListVec &src_ccti_list,
        struct CC_CACongestionSetting& dest_params,
        CCTIEntryListVec& dest_ccti_list,
        bool& enable)
{
    switch (m_ca_user_options[key]) {
    case CA_ENABLE:
        enable = true;
        break;
    case PORT_CONTROL:
        dest_params.Port_Control = src_params.Port_Control;
        break;
    case CA_CONTROL_MAP:
        dest_params.Control_Map = src_params.Control_Map;
        break;
    case CCTI_TIMER:
        for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
            dest_params.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Timer =
            src_params.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Timer;
        }
        break;
    case CCTI_INCR:
        for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
            dest_params.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Increase =
            src_params.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Increase;
        }
        break;
    case TRIGGER_THRESHOLD:
        for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
            dest_params.CACongestionEntryList.CACongestionEntryListElement[sl].Trigger_Threshold =
            src_params.CACongestionEntryList.CACongestionEntryListElement[sl].Trigger_Threshold;
        }
        break;
    case CCTI_MIN:
        for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
            dest_params.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Min =
            src_params.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Min;
        }
        break;
    case CCT:
        dest_ccti_list = src_ccti_list;
        break;
    default:
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "Enterted function with key = %s\n", key.c_str());
        break;
    }
}

/***************/
unsigned int CongestionControlManager::GetNumHosts()
{
    unsigned int num_hosts = cl_qmap_count(&m_p_osm_subn->port_guid_tbl) -
        cl_qmap_count(&m_p_osm_subn->sw_guid_tbl);

    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
            "Found %d hosts in this fabric\n", num_hosts);
    return num_hosts;
}

void CongestionControlManager::NumHostsQuatization(unsigned int& num_hosts)
{
    if (num_hosts <= CC_GENERAL_DEFAULT_MIN_HOSTS) {
        num_hosts = CC_GENERAL_DEFAULT_MIN_HOSTS;
        return;
    }

    // Nearest power of 2 greater than or equal to given number
    num_hosts--;
    num_hosts |= num_hosts >> 1;
    num_hosts |= num_hosts >> 2;
    num_hosts |= num_hosts >> 4;
    num_hosts |= num_hosts >> 8;
    num_hosts |= num_hosts >> 16;

    num_hosts++;
    return;
}

/***************/
void CongestionControlManager::DisableSWParams(struct CC_SwitchCongestionSetting& cc_sw_congestion_setting)
{
    InitCC_SwitchCongestionSetting(cc_sw_congestion_setting);

    cc_sw_congestion_setting.Threshold = 0;
    cc_sw_congestion_setting.Packet_Size = 0xff;
    cc_sw_congestion_setting.Marking_Rate = 0;
}

void CongestionControlManager::DisableCAEntryListElementParams(
        struct CACongestionEntryListElement& cc_ca_cong_entry)
{
    CLEAR_STRUCT(cc_ca_cong_entry);
    cc_ca_cong_entry.CCTI_Increase     = 0;
    cc_ca_cong_entry.CCTI_Min          = CA_CONG_SETTING_DEFAULT_CCTI_MIN;
    cc_ca_cong_entry.Trigger_Threshold = CA_CONG_SETTING_DEFAULT_TRIGGER_THRESHOLD;
    cc_ca_cong_entry.CCTI_Timer        = CA_CONG_SETTING_DEFAULT_CCTI_TIMER;
}



void CongestionControlManager::DisableCAParams(struct CC_CACongestionSetting& cc_ca_congestion_setting)
{

    // HCAs can also be disabled by filling the CCT with 0 
    // Currently it is disabled by set CT_Increase value to 0
    
    cc_ca_congestion_setting.Control_Map  = CA_CONG_SETTING_DEFAULT_CONTROL_MAP;
    cc_ca_congestion_setting.Port_Control = CA_CONG_SETTING_DEFAULT_PORT_CONTROL;
    for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
        DisableCAEntryListElementParams(
            cc_ca_congestion_setting.CACongestionEntryList.CACongestionEntryListElement[sl]);
    }

}


/***************/
bool CongestionControlManager::IsCANodeParamSet(const struct CACCSettingDataBaseNodeEntry &prev_params,
        struct CACCSettingDataBaseNodeEntry& ca_node)
{
    if (ca_node.m_is_congestion_settings_set) {
        ca_node.m_is_congestion_settings_set = CompareCACongestionSetting(ca_node.m_node_params,
                prev_params.m_node_params);
    }

    if (ca_node.m_is_control_table_set) {
        ca_node.m_is_control_table_set = CompareCACongestionControlTableSetting(ca_node.m_node_ccti_list,
                prev_params.m_node_ccti_list);
    }

    return (ca_node.m_is_congestion_settings_set && ca_node.m_is_control_table_set);
}

bool CongestionControlManager::IsSWNodeParamSet(const struct CC_SwitchCongestionSetting &prev_params,
        struct SWCCSettingDataBaseNodeEntry& sw_node)
{
    if (!sw_node.m_is_congestion_settings_set) {
        return false;
    }

    sw_node.m_is_congestion_settings_set = CompareSwitchCongestionSetting(sw_node.m_node_params,
            prev_params);
    return sw_node.m_is_congestion_settings_set;
}

/***************/
void CongestionControlManager::ResetErrorWindow()
{
    if (!m_error_window) {
        // Disable error checking mechanism
        return;
    }
    if (!m_max_errors) {
        // No need to allocate array
        m_num_errors = 0;
        return;
    }
    if (m_p_error_window) {
        delete[] m_p_error_window;
        m_p_error_window = NULL;
    }
    if (!m_p_error_window) {
        m_p_error_window = new struct timeval[m_max_errors];
        for (unsigned int idx = 0; idx < m_max_errors; idx++) {
            m_p_error_window[idx].tv_sec  = 0;
            m_p_error_window[idx].tv_usec = 0;
        }
        m_oldest_error = m_max_errors - 1;
        m_num_errors = 0;
    }
}

/***************/
void CongestionControlManager::UpdateUserOptions()
{
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,"UpdateUserOptions_  \n");

    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator user_ca_node_it;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator user_sw_node_it;
    map<int,CCTIEntryListVec>::iterator ca_ccti_list_it;
    map<int,CCTIEntryListVec>::iterator user_ca_ccti_list_it;
    map<string, bool>::iterator         is_set_flag_it;

    unsigned int previous_num_hosts = m_num_hosts;
    //////////////////////////////////////////////
    // General Options
    //

    m_error_window        = m_master_db.m_error_window;
    m_max_errors          = m_master_db.m_max_errors;
    m_cc_key              = m_master_db.m_cc_key;
    m_num_hosts           = m_master_db.m_num_hosts;
    m_enable              = m_master_db.m_enable;
    m_cc_statistics_cycle = m_master_db.m_cc_statistics_cycle;
    m_cc_log_file_name    = m_master_db.m_cc_log_file_name;
    m_cc_log_file_size    = m_master_db.m_cc_log_file_size;

    // General User Options
    is_set_flag_it = m_user_option_db.m_is_set_flags.begin();
    while (is_set_flag_it != m_user_option_db.m_is_set_flags.end()) {
        SetGeneralOptionsFromUser(is_set_flag_it->first);
        is_set_flag_it++;
    }

    if (!m_num_hosts) {
        // Only if user option wasn't used.
        m_num_hosts = GetNumHosts();
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
            "Number of hosts is %d\n", m_num_hosts);
        NumHostsQuatization(m_num_hosts);
    } else {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Number of hosts is %d\n", m_num_hosts);
    }

    // If number of hosts changes, must recalculate cct list
    bool is_update_cct_list = (previous_num_hosts != m_num_hosts);

    /////////////////////////////////////////////
    // Switch Options
    //

    // Set master default params
    m_sw_cc_setting_db.m_default_params = m_master_db.m_sw_default_params;
    // Set user default for switches
    if (m_enable == false) {
        // Disable default params for switches
        DisableSWParams(m_sw_cc_setting_db.m_default_params);
    }
    else {
        is_set_flag_it = m_user_option_db.m_sw_cc_setting_db.m_is_set_flags.begin();
        while (is_set_flag_it != m_user_option_db.m_sw_cc_setting_db.m_is_set_flags.end()) {
            SetSWOptionsFromUser(is_set_flag_it->first,
                                 m_user_option_db.m_sw_cc_setting_db.m_default_params,
                                 m_sw_cc_setting_db.m_default_params,
                                 m_sw_cc_setting_db.m_enable);
            is_set_flag_it++;
        }
    }

    // Add new switches to db
    user_sw_node_it = m_user_option_db.m_sw_cc_setting_db.m_nodes_db.begin();
    while (user_sw_node_it !=  m_user_option_db.m_sw_cc_setting_db.m_nodes_db.end()) {
        if ((sw_node_it = m_sw_cc_setting_db.m_nodes_db.find(user_sw_node_it->first)) ==
            m_sw_cc_setting_db.m_nodes_db.end()) {
            // Found new switch node , given by user
            SWCCSettingDataBaseNodeEntry
                new_sw_node(user_sw_node_it->first, m_sw_cc_setting_db.m_default_params);
            m_sw_cc_setting_db.m_nodes_db.insert(
              pair<u_int64_t, SWCCSettingDataBaseNodeEntry>(user_sw_node_it->first, new_sw_node));
        }
        user_sw_node_it++;
    }

    struct CC_SwitchCongestionSetting sw_prev_params;
    sw_node_it = m_sw_cc_setting_db.m_nodes_db.begin();
    while (sw_node_it !=  m_sw_cc_setting_db.m_nodes_db.end()) {
        sw_prev_params =  sw_node_it->second.m_node_params;
        // Set default params for this switch
        sw_node_it->second.m_node_params = m_sw_cc_setting_db.m_default_params;
        if (m_enable == false) {
            // When disable, default params will configure disable parameters
            sw_node_it->second.m_enable = false;
            IsSWNodeParamSet(sw_prev_params,sw_node_it->second);
            sw_node_it++;
            continue;
        } else {
            sw_node_it->second.m_error_count = 0; //reset error_count if not in disable mode
        }
        if ((user_sw_node_it = m_user_option_db.m_sw_cc_setting_db.m_nodes_db.find(sw_node_it->first)) !=
            m_user_option_db.m_sw_cc_setting_db.m_nodes_db.end()) {

            sw_node_it->second.m_is_set_flags = user_sw_node_it->second.m_is_set_flags;

            if (user_sw_node_it->second.m_is_set_flags.find("enable") !=
                user_sw_node_it->second.m_is_set_flags.end() &&
                user_sw_node_it->second.m_enable == false) {
                DisableSWParams(sw_node_it->second.m_node_params);
                sw_node_it->second.m_enable = false;
            }
            // Set specific params for this switch
            is_set_flag_it         = user_sw_node_it->second.m_is_set_flags.begin();
            while (is_set_flag_it != user_sw_node_it->second.m_is_set_flags.end()) {
                SetSWOptionsFromUser(is_set_flag_it->first,
                                     user_sw_node_it->second.m_node_params,
                                     sw_node_it->second.m_node_params,
                                     sw_node_it->second.m_enable);
                is_set_flag_it++;
            }
        }
        IsSWNodeParamSet(sw_prev_params,sw_node_it->second);
        sw_node_it++;
    }

    /////////////////////////////////////////////
    // CA Options
    //

    if (is_update_cct_list) {
        ca_ccti_list_it         = m_ca_cc_setting_db.m_default_ccti_list.begin();
        while (ca_ccti_list_it != m_ca_cc_setting_db.m_default_ccti_list.end()) {
            ca_ccti_list_it->second.clear();
            ca_ccti_list_it++;
        }
    }

    // Set master default params
    m_ca_cc_setting_db.m_default_params = m_master_db.m_ca_default_params;
    // Set user default for ca
    if (m_enable == false) {
        // Disable default params for ca
        DisableCAParams(m_ca_cc_setting_db.m_default_params);
    }
    else {
        is_set_flag_it         = m_user_option_db.m_ca_cc_setting_db.m_is_set_flags.begin();
        while (is_set_flag_it != m_user_option_db.m_ca_cc_setting_db.m_is_set_flags.end()) {
            SetCAOptionsFromUser(is_set_flag_it->first,
                                 m_user_option_db.m_ca_cc_setting_db.m_default_params,
                                 m_user_option_db.m_ca_cc_setting_db.m_default_ccti_list[CC_GENERAL_DEFAULT_CCTI_LIST],
                                 m_ca_cc_setting_db.m_default_params,
                                 m_ca_cc_setting_db.m_default_ccti_list[CC_GENERAL_DEFAULT_CCTI_LIST],
                                 m_ca_cc_setting_db.m_enable);
            is_set_flag_it++;
        }
    }

    // Add new ca to db
    user_ca_node_it = m_user_option_db.m_ca_cc_setting_db.m_nodes_db.begin();
    while (user_ca_node_it !=  m_user_option_db.m_ca_cc_setting_db.m_nodes_db.end()) {
        if ((ca_node_it = m_ca_cc_setting_db.m_nodes_db.find(user_ca_node_it->first)) ==
            m_ca_cc_setting_db.m_nodes_db.end()) {
            // Found new ca node , given by user
            CACCSettingDataBaseNodeEntry new_ca_node(
                user_ca_node_it->first, m_ca_cc_setting_db.m_default_params);
            m_ca_cc_setting_db.m_nodes_db.insert(
              pair<u_int64_t, CACCSettingDataBaseNodeEntry>(user_ca_node_it->first, new_ca_node));
        }
        user_ca_node_it++;
    }

    struct CACCSettingDataBaseNodeEntry ca_prev_params;
    ca_node_it = m_ca_cc_setting_db.m_nodes_db.begin();
    while (ca_node_it !=  m_ca_cc_setting_db.m_nodes_db.end()) {
        ca_prev_params = ca_node_it->second;
        bool is_update_ccti_timer = false;

        // Set default params for this ca
        ca_node_it->second.m_node_params = m_ca_cc_setting_db.m_default_params;
        ca_node_it->second.m_is_set_flags = m_user_option_db.m_ca_cc_setting_db.m_is_set_flags;

        if (is_update_cct_list) {
            ca_node_it->second.m_node_ccti_list.clear();
            ca_node_it->second.m_is_control_table_set = false;
        }
        if (m_enable == false) {
            // When disable, default params will configure disable parameters
            ca_node_it->second.m_enable = false;
            IsCANodeParamSet(ca_prev_params, ca_node_it->second);
            ca_node_it++;
            continue;
        } else {
            ca_node_it->second.m_error_count = 0; //reset error_count if not in disable mode
        }

        if ((user_ca_node_it = m_user_option_db.m_ca_cc_setting_db.m_nodes_db.find(ca_node_it->first)) !=
            m_user_option_db.m_ca_cc_setting_db.m_nodes_db.end()) {

            // User configured this node
            //ca_node_it->second.m_is_set_flags = user_ca_node_it->second.m_is_set_flags;
            ca_node_it->second.m_is_set_flags.insert(user_ca_node_it->second.m_is_set_flags.begin(),
                                                     user_ca_node_it->second.m_is_set_flags.end());

            if (user_ca_node_it->second.m_is_set_flags.find("enable") !=
                user_ca_node_it->second.m_is_set_flags.end() &&
                user_ca_node_it->second.m_enable == false) {
                DisableCAParams(ca_node_it->second.m_node_params);
                ca_node_it->second.m_enable = false;
            }
            // Set specific params for this ca
            is_set_flag_it         = user_ca_node_it->second.m_is_set_flags.begin();
            while (is_set_flag_it != user_ca_node_it->second.m_is_set_flags.end()) {
                SetCAOptionsFromUser(is_set_flag_it->first,
                                     user_ca_node_it->second.m_node_params,
                                     user_ca_node_it->second.m_node_ccti_list,
                                     ca_node_it->second.m_node_params,
                                     ca_node_it->second.m_node_ccti_list,
                                     ca_node_it->second.m_enable);
                is_set_flag_it++;
            }

            is_update_ccti_timer =
                (user_ca_node_it->second.m_is_set_flags.find("ccti_timer") !=
                 user_ca_node_it->second.m_is_set_flags.end());
        } else {
            is_update_ccti_timer = (m_user_option_db.m_ca_cc_setting_db.m_is_set_flags.find("ccti_timer") !=
                 m_user_option_db.m_ca_cc_setting_db.m_is_set_flags.end());
        }
        if (!is_update_ccti_timer) {
            // User didn't configure ccti timer.
            // Using previous value. If MTU changed osm_update_needed
            // will reflect it and value will be recalculated.
            for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
                ca_node_it->second.m_node_params.CACongestionEntryList.
                    CACongestionEntryListElement[sl].CCTI_Timer =
                ca_prev_params.m_node_params.CACongestionEntryList.
                    CACongestionEntryListElement[0].CCTI_Timer;
            }
        }
        IsCANodeParamSet(ca_prev_params, ca_node_it->second);

        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "UpdateUserOptions - GUID: " U64H_FMT " LID: %u m_is_congestion_settings_set:%u  m_is_control_table_set:%u \n",
                ca_node_it->second.m_node_info.m_port_guid, 
                ca_node_it->second.m_node_info.m_lid,
                ca_node_it->second.m_is_congestion_settings_set,
                ca_node_it->second.m_is_control_table_set);

        ca_node_it++;
    }

    // Each cycle clear error window
    ResetErrorWindow();
}


/****************************************************/
/**************** public methods ********************/
/****************************************************/
