/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2011, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

#ifndef CC_MGR_TYPES_H_
#define CC_MGR_TYPES_H_

#include <stdlib.h>
#include <sys/types.h>
#include <map>
#include <vector>
#include <string>
using namespace std;

#include <opensm/osm_node.h>

#include <infiniband/ibis/ibis.h>
#include <infiniband/misc/tool_trace/tool_trace.h>


//============================================================================//
/* log defines */
#define OSM_CC_LOG( p_log, level, fmt, ...) {   \
        osm_log( p_log, level, "CC_MGR - " fmt, ## __VA_ARGS__); }
#define OSM_CC_LOG_ENTER( p_log )   \
        osm_log( p_log, OSM_LOG_FUNCS, "CC_MGR - %s: [\n", __func__);
#define OSM_CC_LOG_RETURN( p_log, rc ) {  \
        osm_log( p_log, OSM_LOG_FUNCS, "CC_MGR - %s: ]\n", __func__);   \
        return(rc); }
#define OSM_CC_LOG_RETURN_VOID( p_log ) {   \
        osm_log( p_log, OSM_LOG_FUNCS, "CC_MGR - %s: ]\n", __func__);   \
        return; }

#define CC_MGR_LOG(level, fmt, ...) TT_LOG(TT_LOG_MODULE_CCMGR, level, fmt, ## __VA_ARGS__);

#define CC_GENERAL_DEFAULT_CC_LOG_FILE_NAME "/var/log/ccmgr.log"
#define CC_GENERAL_DEFAULT_CC_LOG_FILE_SIZE 5   // [MB]


//============================================================================//
/* CC infiniband defines */
#define SW_PORT_CONG_SETTING_DEFAULT_CONTROL_TYPE 0

#define SW_CONG_SETTING_DEFAULT_PKT_SIZE 0x8
#define SW_CONG_SETTING_DEFAULT_THRESOLD 0xF
#define SW_CONG_SETTING_DEFAULT_MARKING_RATE 0xA
#define SW_CONG_SETTING_DEFAULT_CONTROL_MAP_MARKING_RATE_IS_VALID 1
#define SW_CONG_SETTING_DEFAULT_CONTROL_MAP_CS_THERSHOLD_AND_CS_RETURN_DELAY_IS_VALID 0
#define SW_CONG_SETTING_DEFAULT_CONTROL_MAP_THRESHOLD_AND_PACKET_SIZE_IS_VALID 1
#define SW_CONG_SETTING_DEFAULT_CONTROL_MAP_CREDIT_MASK_IS_VALID 0
#define SW_CONG_SETTING_DEFAULT_CONTROL_MAP_VICTIM_MASK_IS_VALID 1

#define CA_CONG_SETTING_DEFAULT_CCTI_INCREASE 1
#define CA_CONG_SETTING_DEFAULT_CCTI_TIMER 0
#define CA_CONG_SETTING_DEFAULT_TRIGGER_THRESHOLD 2
#define CA_CONG_SETTING_DEFAULT_CCTI_MIN 0
#define CA_CONG_SETTING_DEFAULT_PORT_CONTROL 0
#define CA_CONG_SETTING_DEFAULT_CONTROL_MAP 0xFFFF
#define CA_CONG_SETTING_DEFAULT_CONTROL_TABLE_CAP 2


//============================================================================//
/* CC general defines */
#define CC_STATISTICS_AGENT_SHOULD_EXIT (m_cc_mgr_request || !m_run_cc_statistics_agent)

#define CONF_INPUT_FILE_NAME "/etc/opensm/cc_mgr.conf"
#define CONF_INPUT_FILE_FLAG "ccmgr --conf_file"

#define CC_GENERAL_DEFAULT_ENABLE               "true"
#define CC_GENERAL_DEFAULT_CC_KEY               0
#define CC_GENERAL_DEFAULT_MTU_SIZE             2048
#define CC_GENERAL_DEFAULT_DIVIDER              3
#define CC_GENERAL_DEFAULT_MAX_MANTICE          0x3FFF
#define CC_GENERAL_DEFAULT_MAX_ERRORS           5
#define CC_GENERAL_DEFAULT_ERROR_WINDOW         5
#define CC_GENERAL_DEFAULT_MIN_HOSTS            32
#define CC_GENERAL_DEFAULT_CCTI_LIST            0
#define CC_GENERAL_DEFAULT_CC_STATISTICS_CYCLE  20

#define SL_MAX_NUM              16
#define CA_MAX_CC_LOG_EVENT     13
#define SW_MAX_CC_LOG_EVENT     15
#define NUM_OF_32_PORTS         2
#define PRINT_PREFIX            "\n\t\t\t"

#ifdef PROD_TREE_ROOT
    #include <md_mod_reg.h>
#endif

const unsigned int ccti_entry_list_elements_num =
    (sizeof(((CCTI_Entry_List*)0)->CCTI_Entry_ListElement) /
     sizeof(CCTI_Entry_ListElement));

//============================================================================//
/* new types - enums */
enum cc_support_t {
    CC_SUPPORT_UNKNOWN, CC_NOT_SUPPORTED, CC_SUPPORTED
};

enum USER_OPTIONS_STR {
    // General Options
    ENABLE,
    CC_KEY,
    NUM_HOSTS,
    MAX_ERRORS,
    ERROR_WINDOW,
    CC_STATISTICS_CYCLE,
    CC_LOG_FILE_NAME,
    CC_LOG_FILE_SIZE,
    // CA Options
    CA_ENABLE,
    PORT_CONTROL,
    CA_CONTROL_MAP,
    CCTI_TIMER,
    CCTI_INCR,
    TRIGGER_THRESHOLD,
    CCTI_MIN,
    CCT,
    // SW Options
    SW_ENABLE,
    THRESHOLD,
    PACKET_SIZE,
    MARKING_RATE,
};


//============================================================================//
// inline init methods
inline static void InitCC_SwitchPortCongestionSetting(
        struct CC_SwitchPortCongestionSetting& cc_sw_port_cong_setting)
{
    CLEAR_STRUCT(cc_sw_port_cong_setting);
}

inline static void InitCC_SwitchCongestionSetting(
        struct CC_SwitchCongestionSetting& cc_sw_cong_setting)
{
    CLEAR_STRUCT(cc_sw_cong_setting);
    cc_sw_cong_setting.Control_Map.Marking_RateIsValid =
    SW_CONG_SETTING_DEFAULT_CONTROL_MAP_MARKING_RATE_IS_VALID;
    cc_sw_cong_setting.Control_Map.CS_ThresholdAndCS_ReturnDelayIsValid =
    SW_CONG_SETTING_DEFAULT_CONTROL_MAP_CS_THERSHOLD_AND_CS_RETURN_DELAY_IS_VALID;
    cc_sw_cong_setting.Control_Map.ThresholdAndPacket_SizeIsValid =
    SW_CONG_SETTING_DEFAULT_CONTROL_MAP_THRESHOLD_AND_PACKET_SIZE_IS_VALID;
    cc_sw_cong_setting.Control_Map.Credit_MaskIsValid =
    SW_CONG_SETTING_DEFAULT_CONTROL_MAP_CREDIT_MASK_IS_VALID;
    cc_sw_cong_setting.Control_Map.Victim_MaskIsValid =
    SW_CONG_SETTING_DEFAULT_CONTROL_MAP_VICTIM_MASK_IS_VALID;

    cc_sw_cong_setting.Marking_Rate = SW_CONG_SETTING_DEFAULT_MARKING_RATE;
    cc_sw_cong_setting.Packet_Size  = SW_CONG_SETTING_DEFAULT_PKT_SIZE;
    cc_sw_cong_setting.Threshold    = SW_CONG_SETTING_DEFAULT_THRESOLD;
}

inline static void InitCC_CACongestionEntryListElement(
        struct CACongestionEntryListElement& cc_ca_cong_entry)
{
    CLEAR_STRUCT(cc_ca_cong_entry);
    cc_ca_cong_entry.CCTI_Increase     = CA_CONG_SETTING_DEFAULT_CCTI_INCREASE;
    cc_ca_cong_entry.CCTI_Min          = CA_CONG_SETTING_DEFAULT_CCTI_MIN;
    cc_ca_cong_entry.Trigger_Threshold = CA_CONG_SETTING_DEFAULT_TRIGGER_THRESHOLD;
    cc_ca_cong_entry.CCTI_Timer        = CA_CONG_SETTING_DEFAULT_CCTI_TIMER;
}

inline static void InitCC_CACongestionSetting(
        struct CC_CACongestionSetting& cc_ca_cong_setting)
{
    CLEAR_STRUCT(cc_ca_cong_setting);
    cc_ca_cong_setting.Control_Map  = CA_CONG_SETTING_DEFAULT_CONTROL_MAP;
    cc_ca_cong_setting.Port_Control = CA_CONG_SETTING_DEFAULT_PORT_CONTROL;
    for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
        InitCC_CACongestionEntryListElement(
            cc_ca_cong_setting.CACongestionEntryList.CACongestionEntryListElement[sl]);
    }
}


//============================================================================//
/* new types - structs */

// CC Statistics Data Base
struct CCStatisticsDataBase {
    u_int64_t    m_num_sw_ports_congested;
};


// CC Node Info
struct CCNodeInfo {
    u_int64_t m_port_guid;
    u_int16_t m_lid;
    osm_node_t  *m_p_osm_node;
    osm_physp_t *m_p_osm_physp;
    u_int8_t  m_sl;
    CCNodeInfo(u_int64_t guid, u_int16_t lid, osm_node_t *p_osm_node = NULL,
            osm_physp_t *p_osm_physp = NULL, u_int8_t sl = 0):
                m_port_guid(guid),m_lid(lid),m_p_osm_node(p_osm_node),m_p_osm_physp(p_osm_physp),m_sl(sl) {}
    CCNodeInfo() : m_port_guid(0),m_lid(0),m_p_osm_node(NULL),m_p_osm_physp(NULL),m_sl(0) {}
};


// SW CC Settings Data Base
struct SWCCSettingDataBaseNodeEntry {
    u_int64_t    m_guid;
    CCNodeInfo   m_node_info;
    bool         m_enable;
    u_int8_t     m_error_count;
    cc_support_t m_cc_support;
    struct CC_SwitchCongestionSetting     m_node_params;
    struct CC_SwitchPortCongestionSetting m_port_params[NUM_OF_32_PORTS];
    map<string, bool>  m_is_set_flags;
    bool               m_osm_update_needed;
    bool               m_is_congestion_settings_set;
    bool               m_is_port_congestion_settings_set;

    SWCCSettingDataBaseNodeEntry(u_int64_t guid):
        m_guid(guid), m_enable(CC_GENERAL_DEFAULT_ENABLE), 
        m_error_count(0), 
        m_cc_support(CC_SUPPORT_UNKNOWN),
        m_osm_update_needed(true),
        m_is_congestion_settings_set(false), m_is_port_congestion_settings_set(false) {
        InitCC_SwitchCongestionSetting(m_node_params);
        for (unsigned int idx = 0; idx < NUM_OF_32_PORTS; idx++) {
            InitCC_SwitchPortCongestionSetting(m_port_params[idx]);
        }
    }

    SWCCSettingDataBaseNodeEntry(u_int64_t guid,
            struct CC_SwitchCongestionSetting node_params):
                m_guid(guid), m_enable(CC_GENERAL_DEFAULT_ENABLE), 
                m_error_count(0), 
                m_cc_support(CC_SUPPORT_UNKNOWN),
                m_node_params(node_params),
                m_osm_update_needed(true),
                m_is_congestion_settings_set(false), m_is_port_congestion_settings_set(false) {
        for (unsigned int idx = 0; idx < NUM_OF_32_PORTS; idx++) {
            InitCC_SwitchPortCongestionSetting(m_port_params[idx]);
        }
    }
};

struct SWCCSettingDataBase {
    bool                                         m_enable;
    struct CC_SwitchCongestionSetting            m_default_params;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry> m_nodes_db;
    map<string, bool>                            m_is_set_flags;

    SWCCSettingDataBase():m_enable(CC_GENERAL_DEFAULT_ENABLE) {
        InitCC_SwitchCongestionSetting(m_default_params);
    }
};


// CA CC Settings Data Base
typedef vector<struct CCTI_Entry_List> CCTIEntryListVec;

struct CACCSettingDataBaseNodeEntry {
    u_int64_t    m_guid;
    CCNodeInfo   m_node_info;
    bool         m_enable;
    u_int8_t     m_error_count;
    cc_support_t m_cc_support;
    unsigned int m_control_table_cap;
    struct CC_CACongestionSetting m_node_params;
    CCTIEntryListVec              m_node_ccti_list;
    map<string, bool>             m_is_set_flags;
    u_int32_t                     m_mtu_size;
    bool                          m_osm_update_needed;
    bool                          m_is_control_table_set;
    bool                          m_is_congestion_settings_set;

    CACCSettingDataBaseNodeEntry(u_int64_t guid, struct CC_CACongestionSetting node_params):
        m_guid(guid), m_enable(CC_GENERAL_DEFAULT_ENABLE), m_error_count(0),
        m_cc_support(CC_SUPPORT_UNKNOWN),
        m_control_table_cap(0),m_node_params(node_params), m_mtu_size(0),
        m_osm_update_needed(true),
        m_is_control_table_set(false), m_is_congestion_settings_set(false) {}

    CACCSettingDataBaseNodeEntry(u_int64_t guid):
        m_guid(guid), m_enable(CC_GENERAL_DEFAULT_ENABLE), m_error_count(0),
        m_cc_support(CC_SUPPORT_UNKNOWN),
        m_control_table_cap(0), m_mtu_size(0),
        m_osm_update_needed(true),
        m_is_control_table_set(false), m_is_congestion_settings_set(false) {
        InitCC_CACongestionSetting(m_node_params);
    }

    CACCSettingDataBaseNodeEntry():
        m_guid(0), m_enable(CC_GENERAL_DEFAULT_ENABLE), m_error_count(0),
        m_cc_support(CC_SUPPORT_UNKNOWN),
        m_control_table_cap(0), m_mtu_size(0),
        m_osm_update_needed(true),
        m_is_control_table_set(false), m_is_congestion_settings_set(false) {
        InitCC_CACongestionSetting(m_node_params);
    }
};

struct CACCSettingDataBase {
    bool                                 m_enable;
    struct CC_CACongestionSetting        m_default_params;
    map<u_int64_t, CACCSettingDataBaseNodeEntry> m_nodes_db;
    map<int, CCTIEntryListVec>           m_default_ccti_list;
    map<string, bool>                    m_is_set_flags;

    CACCSettingDataBase():m_enable(CC_GENERAL_DEFAULT_ENABLE) {
        InitCC_CACongestionSetting(m_default_params);
    }
};


// CC User Option Data Base
struct UserOptionDataBase {
    bool m_enable;
    unsigned int m_num_hosts;
    u_int64_t m_cc_key;
    unsigned int m_max_errors;
    unsigned int m_error_window;
    unsigned int m_cc_statistics_cycle;
    string       m_cc_log_file_name;
    unsigned int m_cc_log_file_size;
    map<string, bool> m_is_set_flags;
    struct SWCCSettingDataBase m_sw_cc_setting_db;
    struct CACCSettingDataBase m_ca_cc_setting_db;
};


// CC Master Data Base
struct MasterDataBase {
    bool m_enable;
    unsigned int m_num_hosts;
    u_int64_t m_cc_key;
    unsigned int m_max_errors;
    unsigned int m_error_window;
    unsigned int m_cc_statistics_cycle;
    string       m_cc_log_file_name;
    unsigned int m_cc_log_file_size;
    struct CC_CACongestionSetting     m_ca_default_params;
    struct CC_SwitchCongestionSetting m_sw_default_params;
};


#endif  /* CC_MGR_TYPES_H_ */

